#include <iostream>
#include "console.h"
#include "vector.h"
using namespace std;

struct VectorPair {
    Vector<int> v1;
    Vector<int> v2;
};

bool partitionable(Vector<int>& nums);

bool partitionable(Vector<int>& rest, int sum1, int sum2);

bool partitionable(Vector<int> &rest, VectorPair &solution);

void partitionable(Vector<int> &rest,
                   VectorPair& singleSolution,
                   Vector<VectorPair>& allSolutions);

void listSubsets(string input, string soFar)
{
    if (input.empty()) {
        cout << "{" << soFar << "}" << endl;
    } else {
        char consider = input[0];
        string rest = input.substr(1);
        listSubsets(rest, soFar + consider);  // explore with
        listSubsets(rest, soFar);             // explore without
    }
}

bool partitionable(Vector<int>& nums) {
    return partitionable(nums, 0, 0);
}

// (1) Find whether a solution exists
bool partitionable(Vector<int>& rest, int sum1, int sum2) {
    if (rest.size() == 0) {
        return sum1 == sum2;
    } else {
        int n = rest[0];
        rest.remove(0);
        bool answer = partitionable(rest, sum1 + n, sum2)
                   || partitionable(rest, sum1, sum2 + n);
        rest.insert(0, n);
        return answer;
    }
}

// (2) Find a single solution
bool partitionable(Vector<int> &rest, VectorPair& solution) {
    if (rest.size() == 0) {
        int sum1 = 0;
        int sum2 = 0;
        for (int val : solution.v1) {
            sum1 += val;
        }

        for (int val : solution.v2) {
            sum2 += val;
        }

        return sum1 == sum2;
    } else {
        int n = rest[0];
        rest.remove(0);

        solution.v1.add(n);

        bool answer1 = partitionable(rest, solution);
        if (answer1) {
            rest.insert(0, n);
            return true;
        }

        solution.v1.remove(solution.v1.size() - 1);

        solution.v2.add(n);

        bool answer2 = partitionable(rest, solution);
        rest.insert(0, n);

        if (answer2) {
            return true;
        }

        solution.v2.remove(solution.v2.size() - 1);

        return false;
    }
}

// (3) find _all_ solutions
void partitionable(Vector<int> &rest,
                   VectorPair& singleSolution,
                   Vector<VectorPair>& allSolutions) {
    if (rest.size() == 0) {
        int sum1 = 0;
        int sum2 = 0;
        for (int val : singleSolution.v1) {
            sum1 += val;
        }

        for (int val : singleSolution.v2) {
            sum2 += val;
        }
        if (sum1 == sum2) {
            allSolutions.add(singleSolution);
        }
    } else {
        int n = rest[0];
        rest.remove(0);

        singleSolution.v1.add(n);

        partitionable(rest, singleSolution, allSolutions);

        singleSolution.v1.remove(singleSolution.v1.size() - 1);

        singleSolution.v2.add(n);

        partitionable(rest, singleSolution, allSolutions);
        rest.insert(0, n);

        singleSolution.v2.remove(singleSolution.v2.size() - 1);
    }
}


int main() {
    listSubsets("ABC", "");
    cout << endl;

    // (1) Is there a solution?
    Vector<int> v1 = {1, 1, 2, 3, 5};
    bool result = partitionable(v1);
    cout << "The Vector " << v1 << " is "
         << (result ? "" : "not ") << "partitionable."
         << endl << endl;

    Vector<int> v2 = {1, 4, 5, 6};
    result = partitionable(v2);
    cout << "The Vector " << v2 << " is "
         << (result ? "" : "not ") << "partitionable."
         << endl << endl;

    // (2) Find a solution
    VectorPair solution = {Vector<int>(), Vector<int>()};
    result = partitionable(v1, solution);
    if (result) {
        cout << v1 << " is partitionable into " << solution.v1 <<
            " and " << solution.v2 << endl;
    }

    solution = {Vector<int>(), Vector<int>()}; // new vectors
    Vector<int> largeVec = {1, 2, 3, 4, 5, 6, 7, 8};
    result = partitionable(largeVec, solution);
    if (result) {
        cout << largeVec << " is partitionable into " << solution.v1
             << " and " << solution.v2 << endl << endl;
    }

    // (3) Find _all_ solutions
    solution = {Vector<int>(), Vector<int>()}; // new vectors
    Vector<VectorPair> allSolutions;
    partitionable(largeVec, solution, allSolutions);

    cout << largeVec << " is partitionable into all of the following:" << endl;
    for (VectorPair singleSolution : allSolutions) {
       cout << singleSolution.v1 << ", "
            << singleSolution.v2 << endl;
    }
    return 0;
}
